// TasksOpcodes.cpp: implementation of client opcodes.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "OpCode.h"
#include "TasksOpcodes.h"
#include "../Files/TaskProcessorFiles.h"
#include "../../SharedFileList.h"
#include "../../SafeFile.h"
#include "../Data/Client.h"

OFFEREDFILE& OFFEREDFILE::operator = (const OFFEREDFILE& stObj)
{
	md4cpy(_Hash, stObj._Hash);
	_ClientID	= stObj._ClientID;
	_ClientPort	= stObj._ClientPort;
	_FileName	= stObj._FileName;
	_FileSize	= stObj._FileSize;
	_FileType	= stObj._FileType;

	return *this;
}


//////////////////////////////////////////////////////////////////////
// Client OpCodes
//////////////////////////////////////////////////////////////////////

void SendEmuleInfoAnswer(SOCKET hSocket, CEmClient_Peer* pClient)
{
	COpCode_EMULEINFOANSWER stMsg;
	stMsg._ClientVersion = CURRENT_VERSION_SHORT;
	stMsg._ProtocolVersion = EMULE_PROTOCOL;
	stMsg._Compression = 1;
	stMsg._UdpVersion = 4;
	stMsg._UdpPort = g_stEngine.Prefs.GetUDPPort();
	stMsg._SourceExchange = 0;//2;
	stMsg._Comments = 1;
	stMsg._ModPlus = CURRENT_PLUS_VERSION;
	stMsg._ExtendedRequest = 2;
	stMsg._ModVersion = _T(PLUS_VERSION_STR);
	stMsg._L2Hac = FILEREASKTIME;
	stMsg._Features = 0; // g_eMuleApp.m_pClientCreditList->CryptoAvailable() ? 3 : 0

	g_stEngine.SendOpCode(hSocket, stMsg, pClient);
}

//////////////////////////////////////////////////////////////////////
// HELLO opcode
bool COpCode_HELLO::ProcessForClient(CEmClient_Peer* pClient)
{
	ASSERT(pClient); // this opcode is processed only in the main context

	// Find client in the list of known clients
	if(pClient->Mule == NULL)
		 pClient->Mule = g_stEngine.ClientList.FindMuleClient(_UserHash);
	// If not found, create it
	if(pClient->Mule == NULL)
	{
		CClientMule* pMule = new CClientMule(_ClientAddr, _UserHash);
		g_stEngine.ClientList.AddClient(pMule);
		pClient->Mule = pMule;
	}
	// Setting parent
	pClient->Mule->Parent = pClient;
	// Connected state
	pClient->Mule->OnConnected();

	// Update client information
	if(_UserName.Valid)		pClient->Mule->UserName = _UserName;
	if(_SoftVersion.Valid)	pClient->Mule->ClientVersion = _SoftVersion;
	if(_UserPort.Valid)		pClient->Mule->ClientPort = _UserPort;
	// to do ModVersion
	if(_UserUDPPort.Valid)	pClient->Mule->ClientUDPPort = _UserUDPPort;
	// to do MiscOptions
	if(_MuleVersion.Valid)	pClient->Mule->MuleVersion = _MuleVersion;

	// Send the answer
	CTask_SayHelloToPeer* pTask = new CTask_SayHelloToPeer();
	if(pTask)
	{
		pTask->SetClient(pClient);
		pTask->m_bManual = true;
		g_stEngine.Sockets.Push(pTask);
	}

	return true;
}

//////////////////////////////////////////////////////////////////////
// HELLO ANSWER
bool COpCode_HELLOANSWER::ProcessForClient(CEmClient_Peer*)
{
	return true;
}

bool COpCode_EMULEINFO::ProcessForClient(CEmClient_Peer* pClient)
{
	ASSERT(pClient); // this opcode is processed only in the main context

	SendEmuleInfoAnswer(m_hSocket, pClient);

	return true;
}

bool COpCode_EMULEINFOANSWER::ProcessForClient(CEmClient_Peer* pClient)
{
	return true;
}

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// Request filename
bool COpCode_REQUESTFILENAME::ProcessForClient(CEmClient_Peer* pClient)
{
	ASSERT(pClient); // this opcode is processed only in the main context

	// Just send requested filename
	CTask_SendRequestedFileName* pTask = new CTask_SendRequestedFileName(_Hash, pClient);
	if(pTask)
		g_stEngine.Files.Push(pTask);

	return true;
}

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// Request filename answer
bool COpCode_REQFILENAMEANSWER::ProcessForClient(CEmClient_Peer* pClient)
{
	return true;
}

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// Start file upload
bool COpCode_STARTUPLOADREQ::ProcessForClient(CEmClient_Peer* pClient)
{
	ASSERT(pClient); // this opcode is processed only in the main context

	// Check that we have that file, and add client to waiting queue
	CTask_UploadReqResult* pTask = new CTask_UploadReqResult(_Hash, pClient);
	if(pTask)
		g_stEngine.Files.Push(pTask);

	return true;
}

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// Accept upload request
bool COpCode_ACCEPTUPLOADREQ::ProcessForClient(CEmClient_Peer*)
{
	return true;
}

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// No such file
bool COpCode_FILEREQANSNOFIL::ProcessForClient(CEmClient_Peer*)
{
	return true;
}

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// Request parts
bool COpCode_REQUESTPARTS::ProcessForClient(CEmClient_Peer* pClient)
{
	ASSERT(pClient); // this opcode is processed only in the main context

	// Prepare and send the blocks
	// Actually, we should remember those blocks' offsets to send them later
	CTask_SendBlock* 
	pTask = new CTask_SendBlock(_Hash, pClient, _Start1, _End1);
	if(pTask)
		g_stEngine.Files.Push(pTask);
	pTask = new CTask_SendBlock(_Hash, pClient, _Start2, _End2);
	if(pTask)
		g_stEngine.Files.Push(pTask);
	pTask = new CTask_SendBlock(_Hash, pClient, _Start3, _End3);
	if(pTask)
		g_stEngine.Files.Push(pTask);

	return true;
}

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// Sending file part
bool COpCode_SENDINGPART::ProcessForClient(CEmClient_Peer*)
{
	return true;
}

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// Get file status
bool COpCode_SETREQFILEID::ProcessForClient(CEmClient_Peer* pClient)
{
	ASSERT(pClient); // this opcode is processed only in the main context

	// Send file status, if we have it
	CTask_SendFileStatus* pTask = new CTask_SendFileStatus(_Hash, pClient);
	if(pTask)
		g_stEngine.Files.Push(pTask);

	return true;
}

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// File status
bool COpCode_FILESTATUS::ProcessForClient(CEmClient_Peer*)
{
	return true;
}

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// Hash request
bool COpCode_HASHSETREQUEST::ProcessForClient(CEmClient_Peer* pClient)
{
	ASSERT(pClient); // this opcode is processed only in the main context

	// Send file hashsets, if we have the file
	CTask_SendHashsets* pTask = new CTask_SendHashsets(_Hash, pClient);
	if(pTask)
		g_stEngine.Files.Push(pTask);

	return true;
}

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// Hash request answer
bool COpCode_HASHSETANSWER::ProcessForClient(CEmClient_Peer*)
{
	return true;
}

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// Offer files
bool COpCode_OFFERFILES::ProcessForClient(CEmClient_Peer*)
{
	return true;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////


///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// (!) can be both we connected to him or he connected to us
// sending HELLO/HELLOANSWER opcode.
bool CTask_SayHelloToPeer::Process()
{
	if (m_pPeer == NULL)
		return true;

	if((m_iMessage == FD_CONNECT && !m_pPeer->m_bFromOutside) || m_bManual)
	{
		COpCode_HELLOANSWER stMsg;
		md4cpy(&stMsg._UserHash, g_stEngine.Prefs.GetUserHash());
		stMsg._ClientAddr.Addr = g_stEngine.GetClientID();
		stMsg._ClientAddr.Port = g_stEngine.Prefs.GetPort();
		stMsg._UserName = g_stEngine.Prefs.GetUserNick();
		stMsg._SoftVersion = EDONKEYVERSION;
		stMsg._UserUDPPort = g_stEngine.Prefs.GetUDPPort();
		stMsg._MiscOptions = 		
			(4						<< 4*6) |	// UDP version
			(1						<< 4*5) |	// Data compression version
			(0/*dwSupportSecIdent*/	<< 4*4) |	// Secure Ident
			(0/*2*/					<< 4*3) |	// Source Exchange
			(2						<< 4*2) |	// Ext. Requests
			(1						<< 4*1) |	// Comments
			(1/*dwNoViewSharedFiles*/<< 1*2);	// No View Shared files
		stMsg._MuleVersion = 
			(PLUS_COMPATIBLECLIENTID << 24) | 
			((CURRENT_PLUS_VERSION & 0x7F00) << 9) |
			((CURRENT_PLUS_VERSION & 0xF0) << 6) | 
			((CURRENT_PLUS_VERSION & 0x7) << 7);
		/*	if(g_stEngine.ServerState.IsConnecting() || g_stEngine.ServerState.IsConnected())
		{
		stMsg._ServerAddr._IPAddr = g_stEngine.ServerState.nAddr;
		stMsg._ServerAddr._Port = g_stEngine.ServerState.nPort;
		}
		else
		{
		stMsg._ServerAddr._IPAddr = 0;
		stMsg._ServerAddr._Port = 0;
		}*/

		g_stEngine.SendOpCode(m_pPeer->m_hSocket, stMsg, m_pPeer);

		if (1/*emule protocol supported*/)
		{
			SendEmuleInfoAnswer(m_pPeer->m_hSocket, m_pPeer);
		}
	}
	else
	{
	}

	return true;
}

enum
{
	TASK_START	= 0,
	TASK_READ	= 1,
	TASK_SEND	= 2
};

//////////////////////////////////////////////////////////////////////
// Send list of shared files
CTask_SendSharedList::CTask_SendSharedList()
{
	m_nState = TASK_START;
}

//////////////////////////////////////////////////////////////////////
bool CTask_SendSharedList::Process()
{
	if(!g_stEngine.ServerState.IsConnected())
		return true;

	switch(m_nState)
	{
	case TASK_START:
		m_nState = TASK_READ;
		g_stEngine.Files.Push(this);
		return false;
	case TASK_READ:
		if(g_stEngine.Files.SharedFiles.m_mapSharedFiles.IsEmpty())
			return true;
		for(POSITION pos = g_stEngine.Files.SharedFiles.m_mapSharedFiles.GetStartPosition(); pos != NULL; )
		{
			CCKey		bufKey;
			CKnownFile*	pKnownFile = NULL;
			g_stEngine.Files.SharedFiles.m_mapSharedFiles.GetNextAssoc(pos, bufKey, pKnownFile);
			if(pKnownFile)
			{
				OFFEREDFILE stFile;
				md4cpy(stFile._Hash, pKnownFile->GetFileHash());

				uint32		dwClientID = 0;
				uint16		uClientPort = 0;

				if (false) // pServer && (pServer->GetTCPFlags() & SRV_TCPFLG_COMPRESSION)
				{
					if (pKnownFile->IsPartFile())
					{
						stFile._ClientID	= 0xFCFCFCFC;
						stFile._ClientPort	= 0xFCFC;
					}
					else
					{
						// Publishing a complete files
						stFile._ClientID	= 0xFBFBFBFB;
						stFile._ClientPort	= 0xFBFB;
					}
				}
				else
				{
					//	If connected and highID, set client ID and port
					if(true) // m_pServerConnect->IsConnected() && !m_pServerConnect->IsLowID()
					{
						stFile._ClientID	= g_stEngine.GetClientID();
						stFile._ClientPort	= g_stEngine.Prefs.GetPort();
					}
				}
				stFile._FileName = pKnownFile->GetFileName();
				stFile._FileSize = pKnownFile->GetFileSize();
				stFile._FileType = pKnownFile->GetFileTypeString();

				//	There's no need to send FT_FILEFORMAT (file extension without ".") here as a server takes it from the file name
				/*if ( ((pServer != NULL) && (pServer->GetTCPFlags() & SRV_TCPFLG_NEWTAGS)) ||
					( (pClient != NULL) && pClient->IsEmuleClient() &&
					( ((pClient->GetClientSoft() == SO_EMULE) && (pClient->GetVersion() >= FORM_CLIENT_VER(0, 42, 7))) ||
					((pClient->GetClientSoft() == SO_PLUS) && (pClient->GetVersion() > FORM_CLIENT_VER(1, 1, 0))) ) ) )
				*/
				m_Files.Add(stFile);
			}
		}
		m_nState = TASK_SEND;
		g_stEngine.Sockets.Push(this);
		return false;
	case TASK_SEND:
		if(m_Files.GetCount() > 0)
		{
			COpCode_OFFERFILES stMsg;
			for(int i = 0; i < m_Files.GetCount(); i++)
			{
				COpCode_OFFERFILES::CStruct_Files stFile;
				md4cpy(&stFile._Hash, m_Files[i]._Hash);
				stFile._ClientAddr.Addr	= m_Files[i]._ClientID;
				stFile._ClientAddr.Port	= m_Files[i]._ClientPort;
				stFile._FileName			= m_Files[i]._FileName;
				stFile._FileSize			= m_Files[i]._FileSize;
				stFile._FileType			= m_Files[i]._FileType;
				stMsg._Files.push_back(stFile);
			}

			SOCKET hServerSocket = g_stEngine.ServerState.hSocket;
			CEmClient* pClient = g_stEngine.Sockets.Lookup(hServerSocket);
			if(pClient)
				g_stEngine.SendOpCode(hServerSocket, stMsg, pClient);
		}
	}
	return true;
}

//////////////////////////////////////////////////////////////////////
// Send requested filename
CTask_SendRequestedFileName::CTask_SendRequestedFileName(HashType Hash, CEmClient* pClient)
{
	m_nState = TASK_READ;
	md4cpy(&m_Hash, &Hash);
	m_pClient = pClient;
}

//////////////////////////////////////////////////////////////////////
bool CTask_SendRequestedFileName::Process()
{
	switch(m_nState)
	{
	case TASK_READ:
		if(!g_stEngine.Files.SharedFiles.m_mapSharedFiles.IsEmpty())
		{
			CKnownFile* pFile = g_stEngine.Files.SharedFiles.GetFileByID(m_Hash.hash);
			if(pFile)
			{
				m_strFileName	= pFile->GetFileName();
				m_nState = TASK_SEND;
				g_stEngine.Sockets.Push(this);
				return false;
			}
		}
		return true;
	case TASK_SEND:
		if(!m_strFileName.IsEmpty() && m_pClient)
		{
			COpCode_REQFILENAMEANSWER stMsg;
			md4cpy(&stMsg._Hash, &m_Hash);
			stMsg._FileName	= m_strFileName;
			g_stEngine.SendOpCode(m_pClient->m_hSocket, stMsg, m_pClient);
		}
	}
	return true;
}

//////////////////////////////////////////////////////////////////////
// Start file upload
CTask_UploadReqResult::CTask_UploadReqResult(HashType Hash, CEmClient_Peer* pClient)
	:m_nState(TASK_READ)
	,m_pFile(NULL)
	,m_pClient(pClient)
{
	md4cpy(&m_Hash, &Hash);
}

//////////////////////////////////////////////////////////////////////
bool CTask_UploadReqResult::Process()
{
	switch(m_nState)
	{
	case TASK_READ:
		// If file does not exists, m_pFile becomes NULL
		if(!g_stEngine.Files.SharedFiles.m_mapSharedFiles.IsEmpty())
			m_pFile = g_stEngine.Files.SharedFiles.GetFileByID(m_Hash.hash);
		m_nState = TASK_SEND;
		g_stEngine.Sockets.Push(this);
		return false;
	case TASK_SEND:
		if(m_pClient)
		{
			if(m_pFile != NULL)
			{
				//COpCode_ACCEPTUPLOADREQ stMsg;
				//g_stEngine.SendOpCode(m_pClient->m_hSocket, stMsg, m_pClient);

				//m_pClient->Mule->ReqFile = m_pFile;
				// Request this file
				if(m_pClient->Mule)
					m_pClient->Mule->RequestFile(m_pFile);
			}
			else
			{
				// We don't have the file
				COpCode_FILEREQANSNOFIL stMsg;
				md4cpy(&stMsg._Hash, &m_Hash);
				g_stEngine.SendOpCode(m_pClient->m_hSocket, stMsg, m_pClient);
			}
		}
	}
	return true;
}

//////////////////////////////////////////////////////////////////////
// Send requested block
CTask_SendBlock::CTask_SendBlock(HashType Hash, CEmClient_Peer* pClient, DWORD dwStart, DWORD dwEnd)
{
	m_nState = TASK_READ;
	md4cpy(&m_Hash, &Hash);
	m_pClient = pClient;
	m_dwStart = dwStart;
	m_dwEnd = dwEnd;
	m_dwToGo = 0;
	m_pFileData = NULL;
}

//////////////////////////////////////////////////////////////////////
bool CTask_SendBlock::Process()
{
	switch(m_nState)
	{
	case TASK_READ:
		if(!g_stEngine.Files.SharedFiles.m_mapSharedFiles.IsEmpty())
		{
			CKnownFile* pFile = g_stEngine.Files.SharedFiles.GetFileByID(m_Hash.hash);
			if(pFile && m_dwStart < pFile->GetFileSize())
			{
				// Check if this file was requested
				if(m_pClient->Mule == NULL || m_pClient->Mule->ReqFile != pFile)
					return true;

				DWORD dwPart = m_dwStart / PARTSIZE;
				DWORD dwStart = dwPart * PARTSIZE;
				DWORD dwEnd = ((pFile->GetFileSize() - dwStart) > PARTSIZE) ? (dwStart + PARTSIZE) : pFile->GetFileSize();
				//	Everyone is limited to a single chunk.
				//	Check that m_dwStartOffset and m_dwEndOffset are in the same chunk and
				//	make sure we don't pass the end of the file.
				if((dwPart != ((m_dwEnd - 1) / PARTSIZE)) || (m_dwEnd > dwEnd))
				{
					//	m_dwEnd goes into the next chunk or is beyond the file end.
					//	Set it to the end of the chunk that m_dwStartOffset is in.
					m_dwEnd = dwEnd;
				}
				//	This can't be a wrapped around request, since it has been limited to a single chunk.
				m_dwToGo = m_dwEnd - m_dwStart;
				//	Create a buffer for file data before we start to work with the file
				m_pFileData = new BYTE[m_dwToGo + 500];
				int	iRc = pFile->ReadFileForUpload(m_dwStart, m_dwToGo, m_pFileData);
				if (iRc >= 0)
				{
					m_nState = TASK_SEND;
					g_stEngine.Sockets.Push(this);
					return false;
				}
				else
					delete[] m_pFileData;
			}
		}
		return true;
	case TASK_SEND:
		if(m_pClient && m_pFileData)
		{
			COpCode_SENDINGPART stMsg;
			md4cpy(&stMsg._Hash, &m_Hash);
			stMsg._StartData = m_dwStart;
			stMsg._EndData = m_dwEnd;
			stMsg._Data = new BYTE[m_dwToGo];
			CopyMemory(stMsg._Data, m_pFileData, m_dwToGo);
			g_stEngine.SendOpCode(m_pClient->m_hSocket, stMsg, m_pClient);
			delete[] m_pFileData;
		}
	}
	return true;
}

//////////////////////////////////////////////////////////////////////
// Send file status
CTask_SendFileStatus::CTask_SendFileStatus(HashType Hash, CEmClient* pClient)
{
	m_nState = TASK_READ;
	md4cpy(&m_Hash, &Hash);
	m_pClient = pClient;
	m_pFileStatus = NULL;
	m_nStatusSize = 0;
}

//////////////////////////////////////////////////////////////////////
bool CTask_SendFileStatus::Process()
{
	switch(m_nState)
	{
	case TASK_READ:
		if(!g_stEngine.Files.SharedFiles.m_mapSharedFiles.IsEmpty())
		{
			CKnownFile* pFile = g_stEngine.Files.SharedFiles.GetFileByID(m_Hash.hash);
			if(pFile)
			{
//				CSafeMemFile packetStream(16);
//				((CPartFile*)pFile)->WritePartStatus(&packetStream);
//				m_nStatusSize = packetStream.GetLength();
//				m_pFileStatus = new BYTE[m_nStatusSize];
//				packetStream.Write(m_pFileStatus, m_nStatusSize);

				m_nStatusSize = pFile->GetED2KPartCount();
				m_pFileStatus = new BYTE[m_nStatusSize];
				for(int i = 0; i < m_nStatusSize; i++)
					m_pFileStatus[i] = 1;

				m_nState = TASK_SEND;
				g_stEngine.Sockets.Push(this);
				return false;
			}
		}
		return true;
	case TASK_SEND:
		if(m_pClient && m_pFileStatus)
		{
			COpCode_FILESTATUS stMsg;
			md4cpy(&stMsg._Hash, &m_Hash);
			for(int i = 0; i < m_nStatusSize; i++)
				stMsg._FileStatus.push_back(m_pFileStatus[i]);
			g_stEngine.SendOpCode(m_pClient->m_hSocket, stMsg, m_pClient);
			delete[] m_pFileStatus;
		}
	}
	return true;
}

//////////////////////////////////////////////////////////////////////
// Send file hashsets
CTask_SendHashsets::CTask_SendHashsets(HashType Hash, CEmClient* pClient)
{
	m_nState = TASK_READ;
	md4cpy(&m_Hash, &Hash);
	m_pClient = pClient;
	m_pFileHashes = NULL;
	m_nHashCount = 0;
}

//////////////////////////////////////////////////////////////////////
bool CTask_SendHashsets::Process()
{
	switch(m_nState)
	{
	case TASK_READ:
		if(!g_stEngine.Files.SharedFiles.m_mapSharedFiles.IsEmpty())
		{
			CKnownFile* pFile = g_stEngine.Files.SharedFiles.GetFileByID(m_Hash.hash);
			if(pFile && pFile->GetHashSet() != NULL)
			{
				m_nHashCount = pFile->GetHashCount();
				m_pFileHashes = new HashType[m_nHashCount];
				CopyMemory(m_pFileHashes, pFile->GetHashSet(), sizeof(HashType)*m_nHashCount);

				m_nState = TASK_SEND;
				g_stEngine.Sockets.Push(this);
				return false;
			}
		}
		return true;
	case TASK_SEND:
		if(m_pClient && m_pFileHashes)
		{
			COpCode_HASHSETANSWER stMsg;
			md4cpy(&stMsg._Hash, &m_Hash);
			for(int i = 0; i < m_nHashCount; i++)
				stMsg._Hashsets.push_back(m_pFileHashes[i]);
			g_stEngine.SendOpCode(m_pClient->m_hSocket, stMsg, m_pClient);
			delete[] m_pFileHashes;
		}
	}
	return true;
}
