#include "StdAfx.h"
#include "resource.h"
#include "otherfunctions.h"
#include "KeyboardShortcut.h"
#include "emule.h"

short GetCodeFromPressedKeys(MSG *pMsg)
{
	short nCode = MASK_ENABLED;

	if (GetAsyncKeyState(VK_MENU) < 0)
		nCode += MASK_ALT;
	if (GetAsyncKeyState(VK_CONTROL) < 0)
		nCode += MASK_CTRL;
	if (GetAsyncKeyState(VK_SHIFT) < 0)
		nCode += MASK_SHIFT;
	if ((GetAsyncKeyState(VK_LWIN) < 0) || (GetAsyncKeyState(VK_RWIN) < 0))
		nCode += MASK_WIN;
	nCode += static_cast<short>(pMsg->wParam);

	return nCode;
}

CString GetStringFromShortcutCode(int iIdx, EnumShortcutStringPrefix ePrefixType, BOOL bReturnNoneIfDisabled)
{
	CString	String = ((ePrefixType == SSP_TAB_PREFIX) ? _T("\t") : ((ePrefixType == SSP_SPACE_PREFIX) ? _T(" ") : _T("")));
	short	nCode = g_eMuleApp.m_pGlobPrefs->GetShortcutCode(iIdx);

	if ((nCode & MASK_ENABLED) != 0) 
	{
		if ((nCode & MASK_ALT) != 0)
			String += _T("Alt+");
		if ((nCode & MASK_CTRL) != 0)
			String += _T("Ctrl+");
		if ((nCode & MASK_SHIFT) != 0)
			String += _T("Shift+");
		if ((nCode & MASK_WIN) != 0)
			String +=_T("Win+");

		uchar byteKey = nCode & MASK_KEY;

		if ((byteKey >= VK_F1) && (byteKey <= VK_F12)) // function keys
			String.AppendFormat(_T("F%u"), byteKey - VK_F1 + 1);
		else if ((byteKey >= 'A') && (byteKey <= 'Z')) // letter keys
			String.AppendChar(byteKey);
		else if ((byteKey >= '0') && (byteKey <= '9')) // number keys
			String.AppendChar(byteKey);
		else
		{
			UINT	dwResStrId = 0;

			if (byteKey == VK_RETURN)	// Enter key
				dwResStrId = IDS_SHORTCUTS_ENTER_KEY;
			else if (byteKey == VK_SPACE) // Spacebar
				dwResStrId = IDS_SHORTCUTS_SPACE_KEY;
			else if (byteKey == VK_TAB) // Tab key
				dwResStrId = IDS_SHORTCUTS_TAB_KEY;
			else if (byteKey == VK_BACK) // Bakcspace key
				dwResStrId = IDS_SHORTCUTS_BACK_KEY;
			else if (byteKey == VK_INSERT) // Insert key
				dwResStrId = IDS_SHORTCUTS_INSERT_KEY;
			else if (byteKey == VK_DELETE) // Delete key
				dwResStrId = IDS_SHORTCUTS_DELETE_KEY;
			else if (byteKey == VK_HOME) // Home key
				dwResStrId = IDS_SHORTCUTS_HOME_KEY;
			else if (byteKey == VK_END) // End key
				dwResStrId = IDS_SHORTCUTS_END_KEY;
			else if (byteKey == VK_PRIOR) // Page Up key
				dwResStrId = IDS_SHORTCUTS_PAGEUP_KEY;
			else if (byteKey == VK_NEXT) // Page Down key
				dwResStrId = IDS_SHORTCUTS_PAGEDOWN_KEY;

			if (dwResStrId != 0)
				String += ::GetResString(dwResStrId);
			else	/*** Should never happen ***/
				String += _T("ERROR");
		}
	}
	else if (bReturnNoneIfDisabled)
		String += GetResString(IDS_SHORTCUTS_NONE);
	else
		return _T("");

	return String;
}


CKeyboardShortcut::CKeyboardShortcut(void)
: m_bEnabled(false)
, m_bAlt(false)
, m_bCtrl(false)
, m_bShift(false)
, m_bWin(false)
, m_byteKey(VK_RETURN)
{
}

CKeyboardShortcut::CKeyboardShortcut(short nCode)
{
	SetCode(nCode);
}

CKeyboardShortcut::~CKeyboardShortcut(void)
{
}

void CKeyboardShortcut::SetCode(short nCode)
{
	m_bEnabled	= static_cast<bool>(nCode & MASK_ENABLED);
	m_bAlt		= static_cast<bool>(nCode & MASK_ALT);
	m_bCtrl		= static_cast<bool>(nCode & MASK_CTRL);
	m_bShift	= static_cast<bool>(nCode & MASK_SHIFT);
	m_bWin		= static_cast<bool>(nCode & MASK_WIN);
	m_byteKey	= static_cast<uchar>(nCode & MASK_KEY);
}
